#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "mpi-utils.h"

#define N 100

#define ODD(n) (n & 1)

int main (int argc, char * argv[]) {
  int i;

  int mpi_rank;
  int mpi_size;
  int partner_rank;

  int message = 0;

  double timers[N];
  double timer;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

  if (ODD(mpi_size)) {
    root_eprintf(mpi_rank,
		 "must be launched with an even number of processes\n");
    FINALIZE_AND_FAIL;
  }

  /* everyone finds their partner */
  partner_rank = ODD(mpi_rank) ? mpi_rank - 1 : mpi_rank + 1;

  /* everyone performs ping pong test with their partner */
  for (i = 0; i < N; i++) {
    timers[i] = MPI_Wtime();

    if (!ODD(mpi_rank)) {
      MPI_Send(&message, 1, MPI_INT, partner_rank, 101,
	       MPI_COMM_WORLD);
      MPI_Recv(&message, 1, MPI_INT, partner_rank, 101,
	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    } else {
      MPI_Recv(&message, 1, MPI_INT, partner_rank, 101,
	       MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      MPI_Send(&message, 1, MPI_INT, partner_rank, 101,
	       MPI_COMM_WORLD);
    }

    timers[i] = MPI_Wtime() - timers[i];
  }

  /* everyone finds timer average locally */
  timer = 0.0;
  for (i = 0; i < N; i++)
    timer += timers[i];
  timer /= N;

  /* timers are added up for root process */
  MPI_Reduce(mpi_rank ? &timer : MPI_IN_PLACE, &timer, 1, MPI_DOUBLE,
	     MPI_SUM, ROOT_PROCESS, MPI_COMM_WORLD);

  /* root process finds global average */
  if (PROCESS_IS_ROOT(mpi_rank)) {
    timer /= mpi_size;
    printf("average communication time: %.16f seconds\n", timer);
  }

  FINALIZE_AND_SUCCEED;
}
