#include <stdbool.h>
#include <stdio.h>

#include <mpi.h>

#include "mpi-utils.h"

static bool check_circuit (unsigned short int input) {
  int i;
  bool b[16];

  for (i = 0; i < 16; i++) {
    b[16 - i-1] = input & 1;
    input >>= 1;
  }

  return (
    ( b[ 0] ||  b[ 1]) && (!b[ 1] || !b[ 3]) && (b[ 2] ||  b[ 3]) &&
    (!b[ 3] || !b[ 4]) && ( b[ 4] || !b[ 5]) && (b[ 5] || !b[ 6]) &&
    ( b[ 5] ||  b[ 6]) && ( b[ 6] || !b[15]) && (b[ 7] || !b[ 8]) &&
    (!b[ 7] || !b[13]) && ( b[ 8] ||  b[ 9]) && (b[ 8] || !b[ 9]) &&
    (!b[ 9] || !b[10]) && ( b[ 9] ||  b[11]) && (b[10] ||  b[11]) &&
    ( b[12] ||  b[13]) && ( b[13] || !b[14]) && (b[14] ||  b[15])
  );
}

static void print_circuit (unsigned short int input) {
  int i;
  bool b[16];

  for (i = 0; i < 16; i++) {
    b[16 - i-1] = input & 1;
    input >>= 1;
  }

  for (i = 0; i < 16; i++)
    putchar(b[i] ? '1' : '0');
  putchar('\n');
  fflush(stdout);
}

int main (int argc, char * argv[]) {
  int i;

  int mpi_rank;
  int mpi_size;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

  for (i = mpi_rank; i < 65535; i += mpi_size)
    if (check_circuit(i))
      print_circuit(i);

  FINALIZE_AND_SUCCEED;
}
