#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "mpi-utils.h"

#define M 10001

int main (int argc, char * argv[]) {
  int i;

  int mpi_rank;
  int mpi_size;

  int l_rank;
  int r_rank;

  int lower;
  int size;

  double * um;
  double * up;
  double * uv;
  double * tmp;

  double x;
  double dx = 1.0/M;
  double t;
  double dt = dx;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);

  /* indices with overlap of 1 in each direction */
  lower = BLOCK_LOW (mpi_rank, mpi_size, M-2);
  size  = BLOCK_SIZE(mpi_rank, mpi_size, M-2) + 2;

  um = malloc(size * sizeof(double));
  up = malloc(size * sizeof(double));
  uv = malloc(size * sizeof(double));

  if (um == NULL || up == NULL || uv == NULL) {
    fprintf(stderr, "%d: allocation error\n", mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  for (i = 0; i < size; i++) {
    x = (i+lower)*dx;
    um[i] = sin(2.0*M_PI*x);
  }

  for (i = 1; i < size-1; i++)
    uv[i] = um[i] + 0.5*(um[i-1] - 2*um[i] + um[i+1]);

  uv[0] = uv[size-1] = 0.0;
  up[0] = up[size-1] = 0.0;

  l_rank = mpi_rank ? mpi_rank - 1 : MPI_PROC_NULL;
  r_rank = mpi_rank < mpi_size-1 ? mpi_rank + 1 : MPI_PROC_NULL;

  t = dt;
  while (t < 1.0) {
    t += dt;

    for (i = 1; i < size-1; i++)
      up[i] = um[i] + uv[i-1] + uv[i+1];

    /* exchange overlapped regions */
    MPI_Sendrecv(&up[1],      1, MPI_DOUBLE, l_rank, 101,
		 &up[size-1], 1, MPI_DOUBLE, r_rank, 101,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    MPI_Sendrecv(&up[size-2], 1, MPI_DOUBLE, r_rank, 102,
		 &up[0],      1, MPI_DOUBLE, l_rank, 102,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    /* up[0] = up[size-1] */
    if (mpi_size > 1) {
      if (mpi_rank == 0)
	MPI_Recv(&up[0], 1, MPI_DOUBLE, mpi_size-1, 103,
		 MPI_COMM_WORLD, MPI_STATUS_IGNORE);

      if (mpi_rank == mpi_size-1)
	MPI_Send(&up[size-1], 1, MPI_DOUBLE, 0, 103,
		 MPI_COMM_WORLD);
    } else {
      up[0] = up[size-1];
    }

    tmp = um;
    um = uv;
    uv = up;
    up = tmp;
  }

  free(uv);
  free(up);
  free(um);

  FINALIZE_AND_SUCCEED;
}
